//________________________________________________________________________________
//
// MonkeyAudio.cpp
//________________________________________________________________________________
//

#include <string.h>
#include <stdio.h>
#include "MonkeyAudio.h"
#include "mp3.h"
#include "debug.h"


CMonkeyAudio::CMonkeyAudio () {
    usedBuffers.clear();
    freeBuffers.clear();
    Init();
}


CMonkeyAudio::~CMonkeyAudio ( ) {
	// Clean up MP3 Player
	StopMP3();
	if( m_pMP3 ) m_pMP3->Release();
	m_pMP3 = NULL;
	ExitMP3();
}


void CMonkeyAudio::Init () {

    // Notes:
    //   _ If preloading is not a necessity, a small number of segments can be allocated
    //   (depending on average number of simultaneous sounds) and the waves could be 
    //   loaded dynamically during gameplay.
    //   _ If preloading is a necessity, this entire array of Sound Files can be scratched
    //   and all sounds can be loaded onto their own segements during initialization. 
    //   I'll leave both in for now.

    int i;

    // Initialize Array of Sound Files
    wszSounds[WLD_RAIN_FLOOR] = L"wld_rain.wav";
    wszSounds[WLD_RAIN_CANOPY] = L"wld_rain_canopy.wav";
    wszSounds[WLD_WIND] = L"wld_wind.wav";
    wszSounds[WLD_WATER] = L"no_sound.wav";           // mnk_captured.wav Cannot be Found
    wszSounds[WLD_UNDERWATER] = L"wld_underwater.wav";
    wszSounds[WLD_INSECTS] = L"wld_snd2.wav";
    wszSounds[WLD_INSECTS2] = L"wld_cricket.wav";
    wszSounds[WLD_BIRDS] = L"wld_snd5.wav";
    wszSounds[WLD_BIRDS2] = L"wld_snd4.wav";
    wszSounds[WLD_OTHER] = L"wld_snd1.wav";
    wszSounds[WLD_THUNDER] = L"wld_thunder.wav";
    wszSounds[WLD_THUNDER2] = L"wld_lightning2.wav";
    wszSounds[WLD_THUNDER3] = L"wld_lightning.wav";
    wszSounds[MNK_RUN_FLOOR] = L"mnk_step_leaves.wav";
    wszSounds[MNK_RUN_FLOOR2] = L"mnk_step_leaves2.wav";
    wszSounds[MNK_RUN_FLOOR3] = L"mnk_step_leaves3.wav";
    wszSounds[MNK_RUN_FLOOR4] = L"mnk_step_leaves4.wav";
    wszSounds[MNK_RUN_TREE] = L"mnk_step_tree.wav";
    wszSounds[MNK_RUN_TREE2] = L"mnk_step_tree2.wav";
    wszSounds[MNK_RUN_TREE3] = L"mnk_step_tree3.wav";
    wszSounds[MNK_RUN_TREE4] = L"mnk_step_tree4.wav";
    wszSounds[MNK_RUN_WATER] = L"mnk_step_water.wav";
    wszSounds[MNK_CLIMB] = L"no_sound.wav";
    wszSounds[MNK_JUMP] = L"mnk_jump.wav";
    wszSounds[MNK_LAND] = L"mnk_land.wav";
    wszSounds[MNK_SUBMERGE] = L"mnk_walk_submerge.wav";
    wszSounds[MNK_DEATH] = L"mnk_death.wav";
    wszSounds[MNK_BIRTH] = L"mnk_scream.wav";
    wszSounds[MNK_SCORE] = L"mnk_victory.wav";
    wszSounds[MNK_DAMAGE] = L"mnk_panic.wav";
    wszSounds[MNK_SENSE] = L"mnk_scream.wav";
    wszSounds[MNK_TAUNT] = L"mnk_laugh.wav";
    wszSounds[NET_MSG_LONG] = L"mnk_call3.wav";
    wszSounds[NET_MSG_SHORT] = L"mnk_call.wav";
    wszSounds[NET_MSG_SHORT2] = L"mnk_call2.wav";
    wszSounds[WPN_SLAP_CHANGE] = L"wpn_bomb_throw.wav";
    wszSounds[WPN_SLAP_FIRE] = L"wpn_bomb_throw.wav";
    wszSounds[WPN_SLAP_CONNECT] = L"wpn_slap_connect2.wav";
    wszSounds[WPN_SHOT_CHANGE] = L"wpn_shot_change.wav";
    wszSounds[WPN_SHOT_RELOAD] = L"wpn_shot_reload.wav";
    wszSounds[WPN_SHOT_FIRE] = L"wpn_shot_fire.wav";
    wszSounds[WPN_SHOT_RICOCHET] = L"wpn_shot_ricochet.wav";
    wszSounds[WPN_BANANA_CHANGE] = L"mnk_call4.wav";
    wszSounds[WPN_BANANA_RELOAD] = L"no_sound.wav";
    wszSounds[WPN_BANANA_FIRE] = L"wpn_banana_launch.wav";
    wszSounds[WPN_BANANA_FLY] = L"wpn_banana_flyby.wav";
    wszSounds[WPN_BANANA_EXPLODE] = L"wpn_bomb_explode.wav";
    wszSounds[WPN_COCO_CHANGE] = L"mnk_call5.wav";
    wszSounds[WPN_COCO_RELOAD] = L"wpn_bomb_light.wav";
    wszSounds[WPN_COCO_IDLE] = L"wpn_bomb_fuse.wav";
    wszSounds[WPN_COCO_FIRE] = L"wpn_bomb_throw.wav";
    wszSounds[WPN_COCO_FLY] = L"wpn_bomb_fuse.wav";
    wszSounds[WPN_COCO_BOUNCE] = L"wpn_bomb_ricochet_tree.wav";
    wszSounds[WPN_COCO_EXPLODE] = L"wpn_bomb_explode.wav";
    wszSounds[WPN_RAIL_CHANGE] = L"mnk_laugh.wav";
    wszSounds[WPN_RAIL_RELOAD] = L"wpn_rail_idle.wav";
    wszSounds[WPN_RAIL_FIRE] = L"wpn_rail_fire.wav";
    wszSounds[NACHO_MAN] = L"no_sound.wav";
    wszSounds[WLD_VICTORY] = L"wld_victory.wav";
    wszSounds[POW_MEDKIT] = L"pow_medkit.wav";

    m_pLoader = NULL;
    m_pPerformance = NULL;
    for ( i=0; i<MAX_SOUNDS; i++ ) {
        m_pSegments[i] = NULL;
    }

    // Initialize Loader and Performance
    CoInitialize( NULL );
    CoCreateInstance( CLSID_DirectMusicLoader, NULL, CLSCTX_INPROC,
                      IID_IDirectMusicLoader8, (void**) &m_pLoader );
    CoCreateInstance( CLSID_DirectMusicPerformance, NULL, CLSCTX_INPROC,
                      IID_IDirectMusicPerformance8, (void**) &m_pPerformance );
    m_pPerformance->InitAudio( NULL, NULL, NULL, DMUS_APATH_DYNAMIC_STEREO, 64,
                               DMUS_AUDIOF_ALL, NULL );

    // Set Loader Search Directory
    // Notes:
    //   _ We don't actually want to do this until we know what the directory structure
    //   for the game is going to be.  However, assuming that the directory structure for
    //   the compiled game will be the same as the source code directory structure, then
    //   we can go ahead and set it.
    WCHAR strSearchPath[MAX_PATH] = L"sound\\snd\\";
    MultiByteToWideChar( CP_ACP, 0, NULL, -1, strSearchPath, MAX_PATH );
    m_pLoader->SetSearchDirectory( GUID_DirectMusicAllTypes, strSearchPath, FALSE );

    // Load All Sounds into m_pSegments Array
    for ( i=0; i<MAX_SOUNDS; i++ ) {
        if( FAILED( m_pLoader->LoadObjectFromFile( CLSID_DirectMusicSegment,
                                                   IID_IDirectMusicSegment8,
                                                   wszSounds[i],
                                                   (LPVOID*) &m_pSegments[i] ) ) ) {
            /*
            MessageBoxPrintf( TEXT( "MonkeyAudio" ),
                              TEXT( "Sound file not found: %i\n" ),
                              i );
            exit(0);
            */
            printf("Unable to open sound file %d\n", i);
            break;
        }
    }

    // Download All Sounds
    for ( i=0; i<MAX_SOUNDS; i++ ) {
        m_pSegments[i]->Download( m_pPerformance );
        m_pSegments[i]->SetRepeats( 0 );
    }

    m_pSegments[WLD_RAIN_FLOOR]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WLD_RAIN_CANOPY]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WLD_WIND]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WLD_WATER]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WLD_UNDERWATER]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_FLOOR]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_FLOOR2]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_FLOOR3]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_FLOOR4]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_TREE]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_TREE2]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_TREE3]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_TREE4]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[MNK_RUN_WATER]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WPN_BANANA_FLY]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WPN_COCO_FLY]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WPN_COCO_IDLE]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
    m_pSegments[WLD_BIRDS2]->SetRepeats( DMUS_SEG_REPEAT_INFINITE );
	m_pSegments[NACHO_MAN]->SetRepeats( 0 );

	pListener = NULL;
    AudioPathBuffer toadd;

    for( i = 0; i < ALLOC_BUFFERS; i++ ) {
        toadd.p3DAudioPath = NULL;
        toadd.pDSB = NULL;
        toadd.pSegState = NULL;
        toadd.owner = NULL;
        m_pPerformance->CreateStandardAudioPath( DMUS_APATH_DYNAMIC_3D, 1, TRUE,
                                                 &toadd.p3DAudioPath );
        if( toadd.p3DAudioPath == NULL ) break;
		if( !i ) {
			// get listener
			toadd.p3DAudioPath->GetObjectInPath(0, DMUS_PATH_PRIMARY_BUFFER,
												0, GUID_All_Objects, 0,
												IID_IDirectSound3DListener,
                                                (void **)&pListener);
		}

        toadd.p3DAudioPath->GetObjectInPath( DMUS_PCHANNEL_ALL, DMUS_PATH_BUFFER, 0,
                                             GUID_NULL, 0, IID_IDirectSound3DBuffer,
                                             (LPVOID*) &(toadd.pDSB) );
		freeBuffers.push_back( toadd );
    }

	if( pListener ) {
		pListener->SetDistanceFactor( 0.005f, DS3D_IMMEDIATE );
		pListener->SetRolloffFactor( 0.0625, DS3D_IMMEDIATE );
		printf("have listener!\n");
	}

	printf("got %d buffers\n", i);
}

void CMonkeyAudio::PlayAtPlayer ( int iSound ) {
    if( m_pPerformance == NULL ) return;
    m_pPerformance->PlaySegmentEx( m_pSegments[iSound], NULL, NULL,
                                   DMUS_SEGF_SECONDARY | DMUS_SEGF_NOINVALIDATE, 0, NULL,
                                   NULL, NULL );
}

AudioPathBuffer CMonkeyAudio::Play ( int iSound, float fx, float fy, float fz ) {
    AudioPathBuffer apb;

    // find a free audio path
    if( freeBuffers.size() == 0 ) return apb;
    if( m_pPerformance == NULL ) return apb;
    apb = freeBuffers.front();
    freeBuffers.pop_front();

    //usedBuffers.push_back( apb );
    m_pPerformance->PlaySegmentEx( m_pSegments[iSound], NULL, NULL,
                                   DMUS_SEGF_SECONDARY | DMUS_SEGF_NOINVALIDATE, 0, &apb.pSegState,
                                   NULL, apb.p3DAudioPath );

    if( apb.pDSB != NULL ) {
        apb.pDSB->SetMinDistance( 100.0f, DS3D_IMMEDIATE );
        apb.pDSB->SetPosition( fx, fy, fz, DS3D_IMMEDIATE );
		apb.p3DAudioPath->SetVolume( 0, 0 );
    }

	usedBuffers.push_back( apb );
    return apb;
#if 0
    // Create an AudioPath
    IDirectSound3DBuffer8* pDSB = NULL;
    AudioPathBuffer apb;
    apb.pDSB = NULL;

    if( !m_p3DAudioPath ) return apb;
    fx /= 100;
    fy /= 100;
    fz /= 100;

    // Create a 3D Sound Buffer
    m_p3DAudioPath->GetObjectInPath( DMUS_PCHANNEL_ALL, DMUS_PATH_BUFFER, 0, GUID_NULL, 0,
        IID_IDirectSound3DBuffer, (LPVOID*) &pDSB );

    if( pDSB ) {
        pDSB->SetMinDistance( 50.0f, DS3D_DEFERRED );
		pDSB->SetVolume( -1 );
        //pDSB->SetConeOutsideVolume( -50, DS3D_DEFERRED );
        pDSB->SetPosition( fx, fy, fz, DS3D_DEFERRED );
    } else printf("unable to get dsb for buf 0\n");

    m_pPerformance->PlaySegmentEx( m_pSegments[iSound], NULL, NULL, DMUS_SEGF_USE_AUDIOPATH | DMUS_SEGF_SECONDARY | DMUS_SEGF_NOINVALIDATE, 0, NULL, NULL, m_p3DAudioPath );

    if( pDSB ) pDSB->Release();

    apb.pDSB = pDSB;
    apb.pDSB = NULL;

    return apb;
#endif
}

void CMonkeyAudio::InitBackground( HWND hWnd ) {
	if ( !InitMP3(hWnd) ) {
		printf( "Error Initializing MP3 Filter.\n" );
		exit(1);
	}

	m_pMP3 = LoadMP3(L"sound/snd/Fighting.mp3");
}

void CMonkeyAudio::PlayBackground ( void ) {
	if ( m_pMP3 )
		PlayMP3( m_pMP3 );
}

void CMonkeyAudio::LoadMus( LPCWSTR wszFilename, bool repeats ) {
	this->repeats = repeats;
	StopMP3();
	if ( m_pMP3 ) m_pMP3->Release();
	m_pMP3 = LoadMP3( wszFilename );
}

void CMonkeyAudio::InitMus( HWND hWnd ) {
	InitMP3( hWnd );
}

void CMonkeyAudio::PlayMus( ) {
	if ( m_pMP3 )
		PlayMP3( m_pMP3 );
}

void CMonkeyAudio::StopMus( ) {
	StopMP3();
}

void CMonkeyAudio::SetVol( long cdb ) {
	SetMP3Volume( cdb );
}

void CMonkeyAudio::SetPosition ( AudioPathBuffer &apb, void *who, float fx, float fy,
                                 float fz ) {
    if( apb.pDSB && who == apb.owner )
        apb.pDSB->SetPosition( fx, fy, fz, DS3D_IMMEDIATE );
}


void CMonkeyAudio::Stop ( AudioPathBuffer &apb, void *who ) {
	if( apb.pDSB && apb.owner == who  && m_pPerformance && apb.pSegState )
	{
		m_pPerformance->Stop( NULL, apb.pSegState, 0, DMUS_SEGF_DEFAULT );
		apb.pSegState->Release();
		apb.pSegState = NULL;
		apb.owner = NULL;
	}
}


void CMonkeyAudio::Update ( ) {
#if 0
    if( pListener ) pListener->CommitDeferredSettings();
#endif
    if( m_pPerformance == NULL ) return;
    list<AudioPathBuffer>::iterator apb;
    list<AudioPathBuffer>::iterator torem;
    AudioPathBuffer apbTemp;
    for( apb = usedBuffers.begin(); apb != usedBuffers.end(); apb++ ) {
        if( S_FALSE == m_pPerformance->IsPlaying(NULL, apb->pSegState) ) {
            apbTemp = *apb;
            torem = apb;
            apb--;
            usedBuffers.erase( torem );
            freeBuffers.push_back( apbTemp );
        }
    }
}
